/*
 * robotModel.c
 * ȈՃfg{bgV~[^
 *
 * Date   : November 14, 2020
 * Author : Kenjiro T. MIURA, Shizuoka University, Japan
 *
 */

/* {bgV~[^ */
#include "GL/glut.h"
#include <math.h>
#include <stdio.h>
// #include <stdlib.h>
#include <string.h>
#include "robotModel.h"

body   Body[MAX_NO_BODY];
face   Face[MAX_NO_FACE];
edge   Edge[MAX_NO_EDGE];
vertex Vertex[MAX_NO_VERTEX];

int No_Body; /* {bg\镔i̐ */

static int base_angle = 0, arm0_angle = 0, arm1_angle = 0;

/* {bǧ`f[^̓ǂݍ */
void readRobotData( char *filename )
{
    int i, j; /* JE^ */
    FILE *fp;
    int no_face, no_edge, no_vertex; /* ,Ő,_̑ */
    int no_f, no_e, no_v; /* X̕i̖,Ő,_̐ */
    int index0, index1, index2, index3;
	errno_t error;

    if ( ( error = fopen_s ( &fp, filename, "r" ) ) != 0 ) {
        printf ( "cannot open %s!\n", filename );
        exit (1);
    }

    no_face = no_edge = no_vertex = 0; /* ̏ */

    fscanf_s ( fp, "%d", &No_Body ); /* {bg̕i̐ */

    for ( i=0; i<No_Body; ++i ) {

        /* iɑ */
        fscanf_s ( fp, "%d", &index0 );
        Body[i].face = &Face[no_face+index0];

        fscanf_s ( fp, "%d", &no_f ); /* ʂ̐̓ǂݍ */

        for ( j=0; j<no_f; ++j ) {
            /* ʂɑŐ, ̖ */
            fscanf_s ( fp, "%d %d", &index0, &index1 );
            Face[no_face+j].edge = &Edge[no_edge+index0];
            if ( index1 == -1 ) Face[no_face+j].next = NULL;
            else Face[no_face+j].next = &Face[no_face+index1];
        }

        fscanf_s ( fp, "%d", &no_e ); /* Ő̐̓ǂݍ */

        for ( j=0; j<no_e; ++j ) { 
            /* Őɑ钸_ */
            fscanf_s ( fp, "%d %d", &index0, &index1 );
            Edge[no_edge+j].start_vertex = &Vertex[no_vertex+index0];
            Edge[no_edge+j].end_vertex = &Vertex[no_vertex+index1];
            /* EŐ,Ő,E, */
            fscanf_s ( fp, "%d %d %d %d", &index0, &index1, &index2, &index3 );
            Edge[no_edge+j].right_edge = &Edge[no_edge+index0];
            Edge[no_edge+j].left_edge  = &Edge[no_edge+index1];
            Edge[no_edge+j].right_face = &Face[no_face+index2];
            Edge[no_edge+j].left_face =  &Face[no_face+index3];
        }
 
        fscanf_s ( fp, "%d", &no_v ); /* _̐̓ǂݍ */

        for ( j=0; j<no_v; ++j ) { /* _f[^̓ǂݍ */
            fscanf_s ( fp, "%f %f %f", &Vertex[no_vertex+j].coord[0],
	                                 &Vertex[no_vertex+j].coord[1],
                                     &Vertex[no_vertex+j].coord[2] );
        }

        no_face += no_f; no_edge += no_e; no_vertex += no_v; /* ̍XV */
    } /* i[v̏I */
} /* readRobotDatȁI */

/* _ƒ_̃f[^ */
static GLdouble eye[3] = { 200.0, 200.0, 100.0 };
static GLdouble center[3] = { 0.0, 0.0, 0.0 };

void addBase(void)
{
    base_angle = (base_angle + 10) % 360;
}

void subtractBase(void)
{
    base_angle = (base_angle - 10) % 360;
}

void addArm0(void)
{
    arm0_angle = (arm0_angle + 10) % 360;
}

void subtractArm0(void)
{
    arm0_angle = (arm0_angle - 10) % 360;
}

void addArm1(void)
{
    arm1_angle = (arm1_angle + 10) % 360;
}

void subtractArm1(void)
{
    arm1_angle = (arm1_angle - 10) % 360;
}

/* ʂɑ钸_̌ */
void getVertexData ( face *pface, int *no_vertex, 
                     GLfloat vertex_data[MAX_NO_VERTEX][3] )
{
    int i; /* JE^ */
    edge *sedge, *nedge; /* ŏ̗Ő,̗Ő */
    face *left_face; /* ̖ */

    *no_vertex = 0; /*  */
    sedge = nedge = pface->edge;
    left_face = nedge->left_face;

    do {
        if( left_face == pface ) {
	        for ( i=0; i<3; ++i ) {
                vertex_data[*no_vertex][i] = nedge->end_vertex->coord[i];
            }
            nedge = nedge->left_edge;
        }
        else {
            for ( i=0; i<3; ++i ) {
                vertex_data[*no_vertex][i] = nedge->start_vertex->coord[i];
            }
            nedge = nedge->right_edge;
        }
        ++*no_vertex;
        left_face = nedge->left_face;
    } while ( nedge != sedge ); /* ̗Őŏ̗ŐɈvI */

} /* getVertexDatȁI */

/* {bg̘r̕` */
void drawArmShading ( body bdy )
{
    int i; /* JE^[ */
    int no_vertex; /* _̐ */
    face *pface; /* ʂւ̃|C^ */
    GLfloat vertex_data[MAX_NO_VERTEX][3]; /* _f[^ */
    GLfloat normal[3]; /* @xNg */
    double norm; /* xNg̃m */
    GLfloat vec0[3], vec1[3];

    pface = bdy.face; /* ̂ɑ */
    
    while ( pface != NULL ) {
        getVertexData( pface, &no_vertex, vertex_data );
        /* @xNg */
        for ( i=0; i<3; ++i ) {
            vec0[i] = vertex_data[2][i] - vertex_data[1][i];
            vec1[i] = vertex_data[0][i] - vertex_data[1][i];
        }
        normal[0] = vec0[1] * vec1[2] - vec0[2] * vec1[1];
        normal[1] = vec0[2] * vec1[0] - vec0[0] * vec1[2];
        normal[2] = vec0[0] * vec1[1] - vec0[1] * vec1[0];
        norm = normal[0] * normal[0] + normal[1] * normal[1] + 
               normal[2] * normal[2];
        norm = sqrt ( norm );
        for ( i=0; i<3; ++i ) normal[i] /= norm;

        glBegin(GL_POLYGON); /* p`̕`Jn */
            glNormal3f( normal[0], normal[1], normal[2] );
            for ( i=0; i<no_vertex; i++ ){
                    glVertex3f(vertex_data[i][0], vertex_data[i][1], 
                               vertex_data[i][2]);
            } /* j[v̏I */
        glEnd(); /* `I */
        pface = pface->next; /* ̖ */
    } /* whilȅI*/
}

/*  XN[NA[BJgJ[𔒂ɃZbgB *
 *  {bg̘rʓh`悷B            @@@@@@@@ */
void ourDisplay (void)
{
    GLfloat material_color0[4] = { 1.0, 0.0, 0.0, 1.0 }; /* gUː */
    GLfloat material_color1[4] = { 0.0, 0.0, 1.0, 1.0 }; /* gUː */
    GLfloat material_color2[4] = { 0.0, 1.0, 0.0, 1.0 }; /* gUː */
    GLfloat material_specular[4] = { 0.2, 0.2, 0.2, 1.0 }; /* ʔː */

    /* obt@[̃NA[*/
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    /* ʔː̃Zbg */
    glMaterialfv(GL_FRONT, GL_SPECULAR, material_specular);

    /* {bg̃x[X̉] */
    glPushMatrix();
        /* {bgx[X̉] */
        glRotatef ( (GLfloat) base_angle, 0.0, 0.0, 1.0 );
        /* bh */
        glMaterialfv(GL_FRONT, GL_DIFFUSE, material_color0);
        glMaterialf(GL_FRONT, GL_SHININESS, 10.0);
        drawArmShading(Body[0]);	/*  {bg̃x[X̕` */

        glPushMatrix();
            /* {bg̘rNo.0̉] */  
            glRotatef ( (GLfloat) arm0_angle, 0.0, -1.0, 0.0 );
            /* u[ */
            glMaterialfv(GL_FRONT, GL_DIFFUSE, material_color1);
            drawArmShading(Body[1]);      /* {bg̘rNo.0̕` */
            /* {bg̘rNo.1̉] */
            glTranslatef ( 40.0, 0.0, 0.0 );
            glRotatef ( (GLfloat) arm1_angle, 0.0, -1.0, 0.0 );
            glTranslatef ( -40.0, 0.0, 0.0 );
            /* u[ */
            glMaterialfv(GL_FRONT, GL_DIFFUSE, material_color2);
            drawArmShading(Body[2]);      /* {bg̘rNo.1̕` */
        glPopMatrix();

    glPopMatrix();
    glFlush();
}

/* OpenGL̏ */
void ourInit (void) 
{
    GLfloat light0_position[] = { 0.0, 0.0, 1.0, 0.0 }; /* Ɩ̈ʒu */
    GLfloat light1_position[] = { 100.0, 100.0, 0.0, 1.0 }; /* Ɩ̈ʒu */

    GLfloat light0_diffuse[] = { 0.8, 0.8, 0.8, 1.0 }; /* gU */
    GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 }; /* gU */

    GLfloat light_specular[] = { 0.3, 0.3, 0.3, 1.0 }; /* ʐ */

    GLfloat lmodel_ambient[] = { 0.1, 0.1, 0.1, 1.0 }; /* ͌ */

    /* ƖNo.0 */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);

    /* ƖNo.1 */
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
    glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);

    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
}

/*  EChEŏɃI[vEChEړ⃊TCYꂽ
 *  Ă΂B */
void ourReshape (int width, int height)
{
    int i; /* JE^[ */
    GLdouble up[3]; /* r[AbvxNg */
    GLdouble vector[3]; /* r[AbvxNgvZpxNg */
    GLdouble norm; /* xNg̃m */

    glViewport (0, 0, width, height);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30.0, (GLfloat) width/(GLfloat) height, 10.0, 1000.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* r[AbvxNgvZB*/
    for ( i=0; i<3; i++ ){
        vector[i] = center[i] - eye[i];
    }
    up[0] = - vector[0] * vector[2];
    up[1] = - vector[1] * vector[2];
    up[2] = vector[0] * vector[0] + vector[1] * vector[1];
    norm = up[0] * up[0] + up[1] * up[1] + up[2] * up[2];
    norm = sqrt ( norm );

    for ( i=0; i<3; ++i ) up[i] /= norm;
    
    /* r[sZbgB*/    
    gluLookAt(eye[0], eye[1], eye[2], center[0], center[1], center[2],
              up[0], up[1], up[2]);
}

/* }EXɂړ */
void ourMouse(int button, int state, int x, int y)
{
    if (state != GLUT_DOWN)
        return;

    switch(button) {
    case GLUT_LEFT_BUTTON:
        addArm1();
        break;
    case GLUT_MIDDLE_BUTTON:
        addBase();
        break;
    case GLUT_RIGHT_BUTTON:
        subtractArm1();
    }
    glutPostRedisplay();
}  

/* L[ɂړ */
void ourKeyboard(unsigned char key, int x, int y)
{
    switch(key) {
    case 27:
        exit(0);
        break;
    case 'l':
        subtractBase();
        break;
    case 'r':
        addBase();
        break;
    case 'u':
        addArm0();
        break;
    case 'd':
        subtractArm0();
        break;
    default:
        break;
    }
    glutPostRedisplay();
}

/*  C[v                            *
 *  lŃEChECjVCY. */
int main(int argc, char **argv) 
{
    char filename[30];

    strcpy_s ( filename, "robot.dat" );

    /* {bg􉽌`f[^̓ǂݍ */
    readRobotData ( filename );

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(400, 400);
    glutInitWindowPosition(100, 100);
    glutCreateWindow("robotModelGlut");
    ourInit();
    glutReshapeFunc(ourReshape);
    glutKeyboardFunc(ourKeyboard);
    glutMouseFunc(ourMouse);
    glutDisplayFunc(ourDisplay);
    glutMainLoop();
    return 0;
}
